﻿using System.Collections;
using System.Collections.Generic;
using UnityEngine;

public class Play : MonoBehaviour {
	public Hand paddle_hand;
	public Hand free_hand;
	public float haptics_duration = 0.01f;  // Seconds for hand pulse.
	public float haptics_strength = 1.0f;   // Strength of hand pulse, 0-1.
	public Robot robot;
	public Balls balls;
	Ball ball_in_play;
	public BallTracking ball_tracking;
	PlayState play_state, player_to_serve, robot_to_serve;
	public Serves serves;
	public TextMesh billboard;
	public List<Bouncer> marked_surfaces;
	int robot_return_count = 0;
	public bool playing_game = false;
	int score_player = 0, score_robot = 0;

	Play() {
		create_play_states ();
	}

	void create_play_states() {
		PlayState pts = new PlayState("player to serve");
		PlayState psp = new PlayState("player serve paddle", "player_paddle");
		PlayState pst = new PlayState("player serve table", "table_near");
		PlayState rt = new PlayState ("robot table", "table_far");
		PlayState rh = new PlayState ("robot hit", "robot_paddle");
		PlayState pt = new PlayState ("player table", "table_near");
		PlayState ph = new PlayState ("player hit", "player_paddle");
		PlayState rts = new PlayState("robot to serve");
		PlayState rsp = new PlayState("robot serve paddle", "robot_paddle");
		PlayState rst = new PlayState("robot serve table", "table_far");
		pts.next_state = psp;
		psp.next_state = pst;
		pst.next_state = rt;
		rt.next_state = rh;
		rh.next_state = pt;
		pt.next_state = ph;
		ph.next_state = rt;
		rts.next_state = rsp;
		rsp.next_state = rst;
		rst.next_state = pt;
		player_to_serve = pts;
		robot_to_serve = rts;
	}

	void Start() {
		hold_ball (free_hand);
	}

	void Update() {
		float delta_t = Time.deltaTime;
		paddle_hand.update_paddle_position (delta_t);
		robot.move_paddle (delta_t);
		balls.move_balls (delta_t);
	}

	public void ball_bounced(Ball b, Bouncer bn) {
		if (b == ball_in_play && play_state != null) {
			PlayState nps = play_state.next_state;
			if (bn.tag == nps.bouncer_tag) {
				play_state = nps;
				// Debug.Log ("Hit " + bn.tag + ", next player state " + play_state.name);
			} else {
				keep_score (play_state);
				// Debug.Log ("Point over, last state " + play_state.name + ", hit surface " + bn.tag);
				play_state = null;
			}
		}
		if (bn.tag == "player_paddle") {
			if (haptics_duration > 0 && haptics_strength > 0)
				paddle_hand.wand.haptic_pulse (haptics_duration, haptics_strength);
			Stroke s = return_ball (b);
			float return_time = (s == null ? 0f : s.contact_time - b.motion.time);
			report_speeds (b, bn, return_time);
			enable_markers();
		}
		if (bn.tag == "robot_paddle") {
			disable_markers();
		}
	}

	Stroke return_ball(Ball b) {
		robot_return_count += 1;
		bool player_serve = (play_state != null && play_state.name == "player serve paddle");
		Stroke s = robot.return_ball (b, player_serve, robot_return_count);
		return s;
	}

	public void hold_ball(Hand hand) {
		Ball b = balls.new_ball ();
		hand.hold_ball (b);
		ball_in_play = b;
		play_state = player_to_serve;
		robot_return_count = 0;
	}

	public bool ball_held(Ball b) {
		return (free_hand.held_ball == b || paddle_hand.held_ball == b);
	}

	public void serve() {
		play_state = robot_to_serve;
		robot_return_count = 0;
		Ball ball = balls.new_ball ();
		ball_in_play = ball;
		Serve s = serves.current_serve ();
		s.serve_ball (ball, this);
	}

	public void start_game() {
		playing_game = true;
		score_player = 0;
		score_robot = 0;
		report_score ();
	}

	void keep_score(PlayState last_state) {
		if (playing_game) {
			string sname = last_state.name;
			if (sname.StartsWith ("robot"))
				score_player += 1;
			else if (sname != "player to serve")
				score_robot += 1;
			report_score ();
		}
	}

	public bool player_serves() {
		int total_score = score_player + score_robot;
		bool ps = (
			total_score < 20 ?
			(((total_score / 2) % 2) == 0) :  // Have not reached deuce.
			((total_score % 2) == 0));        // Have reached deuce.
		return ps;
	}

	void report_score() {
		string nl = System.Environment.NewLine;
		string msg = "Player " + score_player + " Robot " + score_robot;
		if (score_player >= 11 && score_player >= score_robot + 2) {
			msg += nl + nl + "You win!!!";
			playing_game = false;
		} else if (score_robot >= 11 && score_robot >= score_player + 2) {
			msg += nl + nl + "Robot wins";
			playing_game = false;
		} else {
			msg += nl + nl + (player_serves() ? "You serve" : "Robot serves");
			msg += nl + nl + "Game to 11" + (score_player + score_robot >= 20 ? ", win by 2" : "");
		}
		billboard.text = msg;

	}

	void report_speeds(Ball b, Bouncer bn_paddle, float tb_time) {
		string nl = System.Environment.NewLine;
		Vector3 v = bn_paddle.wall_strike_velocity;  // Paddle velocity.
		Vector3 n = paddle_hand.held_paddle.forehand_normal();
		if (bn_paddle.z_top)
			n = -n; // backhand normal
		float vf = Vector3.Dot (v, n);
		float vg = (v - vf * n).magnitude;
		BallState bs = b.motion;
		Vector3 s = bs.angular_velocity * b.radius;
		float ss = -Vector3.Dot (s, Vector3.up);
		Vector3 vb = bs.velocity;
		float ts = Vector3.Dot (s, Vector3.Cross (Vector3.up, vb).normalized);
		float paddle_width = 0.15f;
		Vector3 vbi = bn_paddle.ball_strike_velocity;
		float td = (Vector3.Dot (v, n) * vbi - Vector3.Dot (vbi, n) * v).magnitude;
		float timing = 0.5f * paddle_width * Vector3.Dot (v - vbi, n) / td;
		billboard.text = ("Paddle speed " + v.magnitude.ToString("F1") + " m/s" + 
			", flat " + vf.ToString("F1") + " m/s, graze " + vg.ToString("F1") + " m/s" + nl +
			"Ball speed " + vb.magnitude.ToString ("F1") + " m/s" + nl +
			"Ball spin " + s.magnitude.ToString("F1") + " m/s" +
			", top " + ts.ToString("F1") + " m/s, side " + ss.ToString("F1") + " m/s" + nl +
			"Time to top of bounce " + tb_time.ToString("F3") + " sec" + ", timing " + timing.ToString("F3"));
	}

	void enable_markers() {
		foreach (Bouncer bn in marked_surfaces)
			bn.mark = true;
	}

	void disable_markers() {
		foreach (Bouncer bn in marked_surfaces)
			bn.mark = false;
	}
}

public class PlayState {
	public string name;
	public PlayState next_state;
	public string bouncer_tag;       // Surface to hit to reach this state.
	public PlayState(string name, string bouncer_tag ="") {
		this.name = name;
		this.next_state = null;
		this.bouncer_tag = bouncer_tag;
	}
}