﻿using System.Collections;
using System.Collections.Generic;
using UnityEngine;

public class Robot : MonoBehaviour {
	public Paddle paddle;  			// Robot's paddle
	public Paddle player_paddle;

	public float swing_duration = 0.4f;       // seconds
	public float forehand_coverage = 0.75f;   // Fraction of table covered with forehand
	float backswing_pause_time = 0.0f;  // Time from end of backswing to start of stroke.
	float recovery_time = 0.5f;			// Time from end of swing to reaching recovery paddle position.
	float max_grazing_slope = 1f;

	bool limit_acceleration = true;
	float max_backswing_acceleration = 30f;
	float max_stroke_acceleration = 50f;

	public BallTracking ball_tracking;
	public Table table;  // Use table dimensions

	[System.NonSerialized]  // Prevent changing these values in Unity inspector.
	public string [] patterns = { "Forehand", "Backhand", "Random forehand", "Random", "Diagonal", "Down line", "Elbow", "No return" };
	[System.NonSerialized]
	public int pattern = 0;

	[System.NonSerialized]
	public string [] speeds = { "Slow", "Medium", "Fast" };
	[System.NonSerialized]
	public int speed = 0;

	[System.NonSerialized]
	public string[] spins = { "Top spin", "No spin", "Back spin", "1 back spin, then top spin", "Varied", "Varied back spin" };
	[System.NonSerialized]
	public int spin = 0;

	[System.NonSerialized]
	public string[] movement_speeds = { "Slow", "Medium", "Fast", "Unlimited" };
	[System.NonSerialized]
	public int movement_speed = 0;

	float time_now = 0;
	List<PaddleMotion> motions = new List<PaddleMotion>();

	public Stroke return_ball(Ball b, bool player_serve, int num_hit) {
		if (patterns [pattern] == "No return")
			return null;
		
		BallTrack bt = ball_tracking.predict_ball_flight (b, player_serve);
		//Debug.Log ("Ball " + (bt.in_play ? "in play" : "out of play")
		//	+ " over net " + bt.over_net.Count + " after bounce " + bt.final.Count);
		if (!bt.in_play)
			return null;

		BallState bs = bt.top_of_bounce ();
		// Stroke s = rally.block_return (bs, b);
		Stroke s = aimed_return(bs, b, num_hit);
		set_stroke (s, b.motion.time);
		return s;
	}

	Vector3 placement(BallState bs) {
		// Place on table ball should strike.
		// Table half-width is 0.762 meters.
		float x = 0, z = -1.1f;
		string p = patterns [pattern];
		if (p == "Forehand")
			x = 0.5f;
		else if (p == "Backhand")
			x = -0.5f;
		else if (p == "Random forehand")
			x = Random.Range (0f, 0.6f);
		else if (p == "Random")
			x = Random.Range (-0.6f, 0.6f);
		else if (p == "Diagonal")
			x = (bs.position.x >= 0 ? -0.5f : 0.5f);
		else if (p == "Down line")
			x = (bs.position.x >= 0 ? 0.5f : -0.5f);
		else if (p == "Elbow") {
			Vector3 b = bs.position;
			Vector3 elbow_offset = new Vector3 (0, -0.45f, 0);
			Vector3 e = player_paddle.transform.TransformPoint (elbow_offset);
			float f = (z - e.z) / (b.z - e.z);
			x = (1 - f) * e.x + f * b.x;
		}
		Vector3 table_target = new Vector3 (x, 0.76f, z);
		return table_target;
	}

	float top_spin(int num_hit) {
		// Spin
		float top_spin = 0f;
		float spin_mag = 0f;
		if (speeds [speed] == "Slow")
			spin_mag = 4f;
		else if (speeds [speed] == "Medium")
			spin_mag = 5f;
		else if (speeds [speed] == "Fast")
			spin_mag = 7f;
		if (spins [spin] == "Top spin")
			top_spin = spin_mag;
		else if (spins [spin] == "No spin")
			top_spin = 0;
		else if (spins [spin] == "Back spin")
			top_spin = -spin_mag;
		else if (spins [spin] == "1 back spin, then top spin")
			top_spin = (num_hit == 1 ? -spin_mag : spin_mag);
		else if (spins [spin] == "Varied")
			top_spin = Random.Range (-spin_mag, spin_mag);
		else if (spins [spin] == "Varied back spin")
			top_spin = Random.Range (-spin_mag, 0f);
		return top_spin;
	}

	float aim_above() {
		// How high initial ball velocity should aim above table contact point.
		float ball_fall = 1f;
		if (speeds [speed] == "Slow")
			ball_fall = 1.5f;
		else if (speeds [speed] == "Medium")
			ball_fall = 1.2f;
		else if (speeds [speed] == "Fast")
			ball_fall = 0.9f;
		return ball_fall;
	}

	float speed_per_distance(float topspin) {
		// Ratio of outgoing ball speed to distance to table contact.
		float speed_factor = 2f;
		if (speeds [speed] == "Slow")
			speed_factor = 2.2f + 0.03f * topspin;
		else if (speeds [speed] == "Medium")
			speed_factor = 2.4f + 0.04f * topspin;
		else if (speeds [speed] == "Fast")
			speed_factor = 2.6f + 0.06f * topspin;
		return speed_factor;
	}

	Stroke aimed_return(BallState bs, Ball b, int num_hit) {

		Vector3 bv, bav;
		desired_outgoing_ball(bs, b, num_hit, out bv, out bav);

		Stroke s = paddle_stroke(bs, b, bv, bav);

		// If stroke acceleration is too high, try reducing target ball spin.
		float a = s.pre_swing_acceleration();
		if (a > max_stroke_acceleration) {
			float f = max_stroke_acceleration / a;
			float dv = (1f - Mathf.Sqrt(f)) * s.velocity.magnitude;
			Vector3 bav2 = reduce_spin(bav, bv, 2f*dv/b.radius);
			s = paddle_stroke(bs, b, bv, bav2);
			// Debug.Log("Reduced ball spin to reduce accel " + a + " to " + s.pre_swing_acceleration()
			//	+ " target " + max_stroke_acceleration);
		}

		return s;
	}

	void desired_outgoing_ball(BallState bs, Ball b, int num_hit,
								out Vector3 bv, out Vector3 bav) {
		
		// Determine target ball velocity and spin after hit.
		Vector3 table_target = placement(bs);
		float topspin = top_spin (num_hit);
		float ball_fall = aim_above ();
		float speed_factor = speed_per_distance(topspin);

		Vector3 aim_at = table_target + ball_fall * Vector3.up;
		float flight_distance = (table_target - bs.position).magnitude;
		float ball_speed = speed_factor * flight_distance;
		bv = ball_speed * (aim_at - bs.position).normalized;
		bav = (-topspin / b.radius) * Vector3.Cross(bv, Vector3.up).normalized;
		// Don't try to neutralize side spin since that leads to crazy looking
		// side-motion strokes.
		bav += -Vector3.Dot (bs.angular_velocity, Vector3.up) * Vector3.up;
	}

	Vector3 reduce_spin(Vector3 bav, Vector3 bv, float delta_av) {
		return bav + delta_av * Vector3.Cross(bv, Vector3.up).normalized;
	}

	Stroke paddle_stroke(BallState incoming_ball, Ball b,
						 Vector3 outgoing_ball_velocity,
		                 Vector3 outgoing_ball_angular_velocity) {

		BallState bs = incoming_ball;
		Vector3 bv = outgoing_ball_velocity, bav = outgoing_ball_angular_velocity;

		// Find paddle motion to produce target ball velocity and spin
		Bouncer bn = paddle.forehand_rubber();
		Vector3 pv, pn;
		bn.wall_motion (bs.velocity, bv, bs.angular_velocity, bav,
						b.radius, b.inertia_coef, max_grazing_slope,
						out pv, out pn);

		// Decide which side of paddle to use.
		bool backhand = (bs.position.x > (forehand_coverage - 0.5f)*table.width);

		// Get center of paddle blade by offset from ball position.
		Vector3 paddle_pos = bs.position - pn * (b.radius + 0.5f * paddle.thickness);

		// Shorten swing if needed to avoid hitting table.
		float pre_swing_duration = 0.5f*swing_duration;
		float post_swing_duration = 0.5f*swing_duration;
		float d;
		float ave_speed = 0.5f * pv.magnitude;
		if (table_strike_distance(paddle_pos, -pv, pn, out d) && ave_speed * pre_swing_duration > d)
			pre_swing_duration = d / ave_speed;
		if (table_strike_distance(paddle_pos, pv, pn, out d) && ave_speed * post_swing_duration > d)
			post_swing_duration = d / ave_speed;

		float a = (Mathf.PI / 2) * (pv.magnitude / pre_swing_duration);  // Peak accel for sinusoidal profile.
		// Debug.Log("Paddle speed " + pv.magnitude + " height " + (bs.position.y - table.height)
		// 	+ " time " + pre_swing_duration + " accel " + a
		//	+ " spin " + bav.magnitude * b.radius + " ball speed " + bv.magnitude);

		// Create paddle stroke to produce hit.
		Stroke s = new Stroke (bs.time, paddle_pos, pn, pv, pre_swing_duration, post_swing_duration, backhand);
		return s;
	}

	bool table_strike_distance(Vector3 hit_position, Vector3 paddle_velocity, Vector3 paddle_normal, out float d)
	{
		d = 0;
		float h = table.height;
		if (paddle_velocity.y >= 0)
			return false;
		float ph = Mathf.Sqrt(1.0f - paddle_normal.y * paddle_normal.y) * 0.5f * paddle.width;
		float above_table = hit_position.y - ph - table.height;
		if (above_table <= 0)
			return false;
		float t = above_table / -paddle_velocity.y;
		Vector3 tpos = hit_position + t*paddle_velocity;
		// Add paddle width to edges of table to avoid paddle passing through edge.
		bool over_table = (Mathf.Abs(tpos.x) <= 0.5f * table.width + paddle.width
							&& Mathf.Abs(tpos.z) <= 0.5f * table.length);
		if (over_table) {
			d = t * paddle_velocity.magnitude;
			return true;
		}
		return false;
	}

	Stroke block_return(BallState bs, Ball b) {
		Vector3 spin = b.radius * bs.angular_velocity;
		//Debug.Log ("Top of bounce time " + bs.time + " position " + bs.position.ToString ("F4")
		//	+ " velocity " + bs.velocity.ToString("F4") + " spin " + spin.ToString("F4"));
		Vector3 v = -bs.velocity;
		// Debug.Log ("Return time " + bs.time + " velocity " + v.ToString ("F4"));
		// TODO: Compute top spin and angle paddle to compensate.
		float side_spin = spin.y;
		float top_spin = Vector3.Dot(spin, Vector3.Cross (v, Vector3.up).normalized);
		Vector3 vside = side_spin * Vector3.Cross (v, Vector3.up).normalized;
		Vector3 vtop = top_spin * Vector3.up;
		Vector3 v0 = v + 0.5f * vside;
		Vector3 vp = v0 + 0.5f * vtop;
		vp = Mathf.Max(1.4f - 0.2f*vp.magnitude, 0f) * vp;
		Vector3 np = v0 - 0.3f * top_spin * Vector3.up;
		np = np.normalized;
		bool backhand = (bs.position.x > .38); // Cover 3/4 of table with forehand
		Stroke s = new Stroke (bs.time, bs.position, np, vp, 0.5f*swing_duration, 0.5f*swing_duration, backhand);
		return s;
	}
	// Update paddle position to match hand controller position.
	public void move_paddle(float delta_t) {
		time_now += delta_t;

		if (motions.Count == 0)
			return;

		PaddleMotion m = motions [0];
		bool done = m.move (time_now, paddle);
		if (done)
			motions.Remove (m);
	}

	public void set_stroke(Stroke s, float time_now) {
		this.time_now = time_now;
		float bs_start = time_now;
		float bs_end = Mathf.Max (s.start_time - backswing_pause_time, bs_start);
		Transform t = paddle.transform;
		PaddleMotion backswing = new PaddleMotion();
		backswing.set_motion(bs_start, t.position, t.rotation,
							 bs_end, s.start_position, s.start_rotation);
		limit_backswing_acceleration (backswing, s);
		float r_end = s.end_time + recovery_time;
		Vector3 r_pos = new Vector3 (0f, .85f, 1.5f);
		Quaternion r_rot = (Quaternion.AngleAxis(-30, new Vector3(1f,0f,0f))        // tilt tip down
							* Quaternion.AngleAxis(20, new Vector3(0f,1f,0f))      // tilt tip right, favoring forehand
							* Quaternion.AngleAxis(90, new Vector3(0f,0f,1f))      // point blade normal to horizontal
							* Quaternion.AngleAxis (-90, new Vector3(1f,0f,0f)));  // point tip forward.
		PaddleMotion recovery = new PaddleMotion ();
		recovery.set_motion(s.end_time, s.end_position, s.end_rotation,
							r_end, r_pos, r_rot);

		motions.Clear();
		motions.Add (backswing);
		motions.Add (s);
		motions.Add (recovery);
		/*
		Debug.Log ("Robot acceleration,"
					+ " backswing " + backswing.acceleration ().ToString ("F1")
					+ " stroke " + s.acceleration ().ToString ("F1")
					+ " recovery " + recovery.acceleration ().ToString ("F1"));
		*/
	}

	bool limit_backswing_acceleration(PaddleMotion backswing, Stroke s) {
		
		if (!limit_acceleration)
			return false;
		
		float ba = backswing.acceleration ();
		if (ba <= max_backswing_acceleration)
			return false;

		// Lengthen backswing time to stay within maximum acceleration.
		float f = Mathf.Sqrt (ba / max_backswing_acceleration);
		float bs_start = backswing.start_time;
		float bs_end = bs_start + (backswing.end_time - bs_start) * f;
		backswing.set_motion(bs_start, backswing.start_position, backswing.start_rotation,
							 bs_end, s.start_position, s.start_rotation);
		// Debug.Log("Reduced backswing accel " + backswing.acceleration() + " max " + max_backswing_acceleration);

		// If stroke begins before backswing ends then shorten stroke.
		if (bs_end > s.start_time) {
			float fs = Mathf.Sqrt (s.pre_swing_acceleration () / max_stroke_acceleration);
			float min_dur = fs * s.pre_swing_duration;
			float new_dur = s.pre_swing_duration - (bs_end - s.start_time);
			new_dur = Mathf.Max (new_dur, min_dur);
			s.change_pre_swing_duration (new_dur);
			// Fix backswing end position for shorter stroke.
			backswing.set_motion(bs_start, backswing.start_position, backswing.start_rotation,
								bs_end, s.start_position, s.start_rotation);
			// Debug.Log ("Shortened stroke duration to " + tmin.ToString ("F2"));
			// Debug.Log("new backswing accel " + backswing.acceleration() + " new stroke accel " + s.pre_swing_acceleration());
			// If there still not time for shorter maximum acceleration stroke
			// then delay the stroke. This will often cause a ball miss.
			if (bs_end > s.start_time) {
				float dt = bs_end - s.start_time;
				s.start_time += dt;
				s.end_time += dt;
				// Debug.Log ("Delayed stroke " + dt.ToString ("F2"));
			}
		}
		return true;
	}

	public void set_movement_speed() {
		string s = movement_speeds[movement_speed];
		if (s == "Slow") {
			limit_acceleration = true;
			max_backswing_acceleration = 30f;
			max_stroke_acceleration = 50f;
		} else if (s == "Medium") {	
			limit_acceleration = true;
			max_backswing_acceleration = 50f;
			max_stroke_acceleration = 100f;
		} else if (s == "Fast") {
			limit_acceleration = true;
			max_backswing_acceleration = 70f;
			max_stroke_acceleration = 150f;
		} else if (s == "Unlimited") {
			limit_acceleration = false;
		}
	}
}
