﻿using System.Collections;
using System.Collections.Generic;
using UnityEngine;
using Valve.VR;

[RequireComponent(typeof(SteamVR_TrackedObject))]
public class Wand : MonoBehaviour {
	
	public SteamVR_TrackedObject hand;
	public bool right, left;
	public Transform tracking_space;  // Coordinate system for local hand controller positions.
	public float latency = 0f;  // Time in seconds that controller lags frame redraw.
	// Display warning message if controller maximum acceleration exceeded.
	public Warning warning;      // For displaying warning messages.
	float last_hs = 0f;  // Last hand speed, for debugging
	float last_ac = 0f;   // Last accel magnitude, for debugging

	bool trigger_released = true;
	bool grip_released = true;
	bool thumbstick_released = true;

	ulong OculusTouch_ButtonA = (1ul << 7);   // SteamVR bit for Oculus Touch controller button A (right hand) or X (left hand).
	ulong OculusTouch_ButtonB = (1ul << 1);   // SteamVR bit for Oculus Touch controller button A (right hand) or X (left hand).
	                                          //   same as SteamVR_Controller.ButtonMask.ApplicationMenu

	SteamVR_Controller.Device device() {
		int i = (int)hand.index;
		if (i == -1)
			return null;
		//Debug.Log ("Hand controller index " + i);
		var dev = SteamVR_Controller.Input(i);
		return dev;
	}

	TrackedDevicePose_t[] poses = new TrackedDevicePose_t[OpenVR.k_unMaxTrackedDeviceCount];
	TrackedDevicePose_t[] last_poses = new TrackedDevicePose_t[OpenVR.k_unMaxTrackedDeviceCount];
	Vector3 prev_hv;

	public void wand_motion(out Vector3 position, out Quaternion rotation,
		out Vector3 velocity, out Vector3 angular_velocity,
		out Vector3 acceleration) {

		int i = (int)hand.index;
		var compositor = OpenVR.Compositor;
		if (i == -1 || compositor == null) {
			position = velocity = angular_velocity = acceleration = Vector3.zero;
			rotation = Quaternion.identity;
			return;
		}

		// Get predicted poses for previous and next rendered frames.
		compositor.GetLastPoses(poses, last_poses);

		var pose = last_poses [i];
		// TODO: Check if poses are valid using pose.bPoseIsValid
		var t = new SteamVR_Utils.RigidTransform(pose.mDeviceToAbsoluteTracking);
		Vector3 hp = t.pos;
		Quaternion hr = t.rot;

		// Sign changes for velocity are copied form SteamVR_Controller.Device class.
		Vector3 hv = new Vector3(pose.vVelocity.v0, pose.vVelocity.v1, -pose.vVelocity.v2);
		Vector3 hav = new Vector3 (-pose.vAngularVelocity.v0, -pose.vAngularVelocity.v1, pose.vAngularVelocity.v2);

		// Approximate acceleration from two velocity measurements.
		// With Samsung Odyssey Windows Mixed Reality headset and SteamVR
		// the previous velocity and next velocity are always identical (Dec 31, 2018).
		//var prev_pose = poses [i];
		//Vector3 prev_hv = new Vector3(prev_pose.vVelocity.v0, prev_pose.vVelocity.v1, -prev_pose.vVelocity.v2);
		// TODO: Don't hard code time step.
		float time_step = 1.0f / 90.0f;
		Vector3 ha = (hv - prev_hv) / time_step;
		prev_hv = hv;

		// Transform from tracking_space coordinates (camera) to world coordinates.
		position = tracking_space.TransformPoint (hp);
		rotation = tracking_space.rotation * hr;
		velocity = tracking_space.TransformVector (hv);
		angular_velocity = tracking_space.TransformVector (hav);
		acceleration = tracking_space.TransformVector (ha);
	}

	public void wand_motion1(out Vector3 position, out Quaternion rotation,
							out Vector3 velocity, out Vector3 angular_velocity,
							out Vector3 acceleration) {
		// Object position is in hand coordinates.
		// Outputs are in world coordinates.
		var dev = device();
		if (dev == null) {
			position = velocity = angular_velocity = acceleration = Vector3.zero;
			rotation = Quaternion.identity;
			return;
		}
		dev.Update();
		//Vector3 hp = dev.transform.pos;
		//Quaternion hr = dev.transform.rot;
		Vector3 hp = hand.transform.localPosition;
		Quaternion hr = hand.transform.localRotation;
		Vector3 hv = dev.velocity;
		//Vector3 ha = device.acceleration;
		Vector3 ha = Vector3.zero; // TODO: SteamVR does not have acceleration.
		Vector3 hav = dev.angularVelocity;
		// TODO: Report bug in OVRInput angular velocity is not in tracking frame.
		//Vector3 hav = hr * hav1;

		// Issue warning if hand velocity and accel zero which happens
		// when 16g accleration exceeded with oculus touch.
		//check_imu_out_of_range (hv, ha);

		// Transform from tracking_space coordinates to world coordinates.
		position = tracking_space.TransformPoint (hp);
		rotation = tracking_space.rotation * hr;
		velocity = tracking_space.TransformVector (hv);
		angular_velocity = tracking_space.TransformVector (hav);
		acceleration = tracking_space.TransformVector (ha);
	}

	public void object_motion(Vector3 object_position, out Vector3 position, out Quaternion rotation,
								out Vector3 velocity, out Vector3 acceleration) {
		// Object position is in hand coordinates.
		// Outputs are in world coordinates.
		Vector3 p, va;
		Quaternion r;
		wand_motion(out p, out r, out velocity, out va, out acceleration);
		position = p + r * object_position;
		rotation = r;
	}

	// Returned hand position, rotation, velocity, angular velocity and
	// acceleration are in tracking_space coordinates.
	public void predict_hand_motion(out Vector3 hp, out Quaternion hr,
									out Vector3 hv, out Vector3 hav, out Vector3 ha) {
		wand_motion (out hp, out hr, out hv, out hav, out ha);

		// Predict controller position at the next frame draw time.
		// Unfortunately the Unity Oculus 1.18.1 API does not provide a way to predict
		// the pose at the next redraw, so I have to compute it myself.
		if (latency != 0) {
			// Extrapolate hand position.
			float t = latency;
			hp += t * hv + (0.5f * t * t) * ha;
			float ad = Mathf.Rad2Deg * t * hav.magnitude;
			Vector3 axis = hav.normalized;
			hr = Quaternion.AngleAxis (ad, axis) * hr;
			// TODO: Predict angular velocity using angular acceleration.
		}
	}

	// Look for hand velocity that happens when Oculus Touch controller inertial measurement
	// unit hardware exceeds 16g acceleration limit.
	void check_imu_out_of_range(Vector3 hv, Vector3 ha) {
		if (hv.magnitude == 0 && last_hs >= 5) {
			warning.warn ("Acceleration out of range, last hand speed " + last_hs + " and accel " + last_ac);
		}
		last_hs = hv.magnitude;
		last_ac = ha.magnitude;
	}

	public bool trigger_press() {
		var dev = device ();
		if (dev == null)
			return false;
		float pressed = dev.GetAxis (EVRButtonId.k_EButton_Axis1).x;
		if (trigger_released && pressed > .5) {
			trigger_released = false;
			return true;
		} else if (pressed < .3) {
			trigger_released = true;
		}
		return false;
	}

	public bool grip_press() {
		var dev = device ();
		if (dev == null)
			return false;
		float pressed = dev.GetAxis (EVRButtonId.k_EButton_Axis2).x;
		if (grip_released && pressed > .5) {
			grip_released = false;
			return true;
		} else if (pressed < .3) {
			grip_released = true;
		}
		return false;
	}

	public Vector2 direction_click() {
		var dev = device ();
		if (dev == null)
			return Vector2.zero;
		Vector2 ts = dev.GetAxis (EVRButtonId.k_EButton_SteamVR_Touchpad); // OVRInput.Axis2D.PrimaryThumbstick
		float p = Mathf.Max (Mathf.Abs (ts.x), Mathf.Abs (ts.y));
		if (p < .3)
			thumbstick_released = true;
		else if (p > .5 && thumbstick_released) {
			thumbstick_released = false;
			if (ts.x > 0.5f)
				return new Vector2 (1, 0);
			else if (ts.x < -0.5f)
				return new Vector2 (-1, 0);
			else if (ts.y > 0.5f)
				return new Vector2 (0, 1);
			else if (ts.y < -0.5f)
				return new Vector2 (0, -1);
		}
		return Vector2.zero;
	}

	public bool button1_pressed() {
		var dev = device ();
		if (dev == null)
			return false;		
		bool p = dev.GetPressDown (OculusTouch_ButtonA); // OVRInput.RawButton.A
		return p;
	}

	public bool button2_pressed() {
		var dev = device ();
		if (dev == null)
			return false;
		bool p = dev.GetPressDown (OculusTouch_ButtonB);
		return p;
	}

	public void haptic_pulse(float duration, float strength) {
		var dev = device ();
		if (dev == null)
			return;
		StartCoroutine (haptic_pulses (dev, duration, strength));
		//dev.TriggerHapticPulse (durationMicroSeconds);
	}

	IEnumerator haptic_pulses(SteamVR_Controller.Device device, float duration, float strength) {
		ushort pulse_microsec = (ushort) (3999 * strength); // Max pulse duration is 4 milliseconds.
		for (float t = 0; t < duration; t += Time.deltaTime) {
			device.TriggerHapticPulse (pulse_microsec);
			yield return new WaitForSeconds(0.005f);  // Can only trigger pulse every 5 msec according to SteamVR docs.
		}
	}
}
